@@ -1,8 +1,9 @@
 /*
- * Copyright © 2009 CNRS
- * Copyright © 2009-2022 Inria.  All rights reserved.
+ * Copyright © 2009,2024 CNRS
+ * Copyright © 2009-2024 Inria.  All rights reserved.
  * Copyright © 2009-2011 Université Bordeaux
  * Copyright © 2009-2010 Cisco Systems, Inc.  All rights reserved.
+ * Copyright © 2023 Université de Reims Champagne-Ardenne.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -54,8 +55,14 @@ void usage(const char *callname __hwloc_attribute_unused, FILE *where)
   fprintf(where, "  -n --nodeset              Manipulate nodesets instead of cpusets\n");
   fprintf(where, "  --ni --nodeset-input      Manipulate nodesets instead of cpusets for inputs\n");
   fprintf(where, "  --no --nodeset-output     Manipulate nodesets instead of cpusets for outputs\n");
+  fprintf(where, "  --oo --object-output      Report objects instead of object indexes\n");
   fprintf(where, "  --sep <sep>               Use separator <sep> in the output\n");
-  fprintf(where, "  --taskset                 Use taskset-specific format when displaying cpuset strings\n");
+  fprintf(where, "  --cpuset-input-format <hwloc|list|taskset>\n"
+                 "  --cif <hwloc|list|taskset>\n"
+                 "                            Change the format of cpuset inputs\n");
+  fprintf(where, "  --cpuset-output-format <hwloc|list|taskset|systemd-dbus-api>\n"
+                 "  --cof <hwloc|list|taskset>\n"
+                 "                            Change the format of cpuset outputs\n");
   fprintf(where, "  --single                  Singlify the output to a single CPU\n");
   fprintf(where, "Miscellaneous options:\n");
   fprintf(where, "  -q --quiet                Hide non-fatal error messages\n");
@@ -69,19 +76,19 @@ static int logicali = 1;
 static int logicalo = 1;
 static int nodeseti = 0;
 static int nodeseto = 0;
-static int numberofdepth = -1;
-static union hwloc_obj_attr_u numberofattr;
-static int intersectdepth = -1;
-static union hwloc_obj_attr_u intersectattr;
-static int hiernblevels = 0;
-static int *hierdepth = NULL;
+static int objecto = 0;
+static struct hwloc_calc_level numberof;
+static struct hwloc_calc_level intersect;
+static int hiernblevels;
+static struct hwloc_calc_level *hierlevels;
 static int local_numanodes = 0;
 static unsigned long local_numanode_flags = HWLOC_LOCAL_NUMANODE_FLAG_SMALLER_LOCALITY | HWLOC_LOCAL_NUMANODE_FLAG_LARGER_LOCALITY;
 static hwloc_memattr_id_t best_memattr_id = (hwloc_memattr_id_t) -1;
+static unsigned long best_node_flags = 0;
 static int showobjs = 0;
 static int no_smt = -1;
 static int singlify = 0;
-static int taskset = 0;
+static enum hwloc_utils_cpuset_format_e cpuset_output_format = HWLOC_UTILS_CPUSET_FORMAT_HWLOC;
 static hwloc_bitmap_t cpukind_cpuset = NULL;
 
 static int
@@ -116,12 +123,14 @@ hwloc_calc_hierarch_output(hwloc_topology_t topology, const char *prefix, const
   hwloc_obj_t obj, prev = NULL;
   unsigned logi = 0;
   int first = 1;
-  while ((obj = hwloc_get_next_obj_covering_cpuset_by_depth(topology, root->cpuset, hierdepth[level], prev)) != NULL) {
+  while ((obj = hwloc_get_next_obj_covering_cpuset_by_depth(topology, root->cpuset, hierlevels[level].depth, prev)) != NULL) {
     char string[256];
     char type[32];
     unsigned idx = logicalo ? logi : obj->os_index;
     if (!hwloc_bitmap_intersects(set, obj->cpuset))
      goto next;
+    if (hwloc_calc_check_object_filtered(obj, &hierlevels[level]))
+      goto next;
     hwloc_obj_type_snprintf(type, sizeof(type), obj, 1);
     if (idx == (unsigned)-1)
       snprintf(string, sizeof(string), "%s%s%s:-1", prefix, level ? "." : "", type);
@@ -187,30 +196,31 @@ hwloc_calc_output(hwloc_topology_t topology, const char *sep, hwloc_bitmap_t set
     }
     printf("\n");
     hwloc_bitmap_free(remaining);
-  } else if (numberofdepth != -1) {
+  } else if (numberof.depth != -1) {
     unsigned nb = 0;
     hwloc_obj_t obj = NULL;
-    while ((obj = hwloc_calc_get_next_obj_covering_set_by_depth(topology, set, nodeseto, numberofdepth, obj)) != NULL) {
-      if (numberofdepth == HWLOC_TYPE_DEPTH_OS_DEVICE
-          && numberofattr.osdev.type != (hwloc_obj_osdev_type_t) -1
-          && numberofattr.osdev.type != obj->attr->osdev.type)
+    while ((obj = hwloc_calc_get_next_obj_covering_set_by_depth(topology, set, nodeseto, numberof.depth, obj)) != NULL) {
+      if (hwloc_calc_check_object_filtered(obj, &numberof))
         continue;
       nb++;
     }
     printf("%u\n", nb);
-  } else if (intersectdepth != -1) {
+  } else if (intersect.depth != -1) {
     hwloc_obj_t obj = NULL;
     int first = 1;
     if (!sep)
       sep = ",";
-    while ((obj = hwloc_calc_get_next_obj_covering_set_by_depth(topology, set, nodeseto, intersectdepth, obj)) != NULL) {
+    while ((obj = hwloc_calc_get_next_obj_covering_set_by_depth(topology, set, nodeseto, intersect.depth, obj)) != NULL) {
       unsigned idx;
-      if (intersectdepth == HWLOC_TYPE_DEPTH_OS_DEVICE
-          && intersectattr.osdev.type != (hwloc_obj_osdev_type_t) -1
-          && intersectattr.osdev.type != obj->attr->osdev.type)
+      if (hwloc_calc_check_object_filtered(obj, &intersect))
         continue;
       if (!first)
 	printf("%s", sep);
+      if (objecto) {
+        char types[64];
+        hwloc_obj_type_snprintf(types, sizeof(types), obj, 0);
+        printf("%s:", types);
+      }
       idx = logicalo ? obj->logical_index : obj->os_index;
       if (idx == (unsigned)-1)
         printf("-1");
@@ -228,124 +238,77 @@ hwloc_calc_output(hwloc_topology_t topology, const char *sep, hwloc_bitmap_t set
   } else if (local_numanodes) {
     unsigned nrnodes;
     hwloc_obj_t *nodes;
+    hwloc_nodeset_t nodeset = hwloc_bitmap_alloc_full(); /* show all nodes by default */
     nrnodes = hwloc_bitmap_weight(hwloc_topology_get_topology_nodeset(topology));
     nodes = malloc(nrnodes * sizeof(*nodes));
-    if (nodes) {
+    if (nodeset && nodes) {
       int err;
       struct hwloc_location loc;
       loc.type = HWLOC_LOCATION_TYPE_CPUSET;
       loc.location.cpuset = set;
       err = hwloc_get_local_numanode_objs(topology, &loc, &nrnodes, nodes, local_numanode_flags);
       if (!err) {
-        unsigned i;
+        unsigned i, first = 1;
         if (best_memattr_id != (hwloc_memattr_id_t) -1) {
-          int best = hwloc_utils_get_best_node_in_array_by_memattr(topology, best_memattr_id, nrnodes, nodes, &loc);
-          if (best == -1) {
-            /* no perf info found, report nothing */
-            nrnodes = 0;
-          } else {
-            /* only report the best nodes */
-            nodes[0] = nodes[best];
-            nrnodes = 1;
-          }
+          err = hwloc_utils_get_best_node_in_array_by_memattr(topology, best_memattr_id, nrnodes, nodes, &loc, best_node_flags, nodeset);
+          /* on error, nodeset is zeroed, and we report nothing below (except if default flag is set) */
         }
         if (!sep)
           sep = ",";
         for(i=0; i<nrnodes; i++) {
           char type[64];
           unsigned idx;
+          if (!hwloc_bitmap_isset(nodeset, nodes[i]->os_index))
+            continue;
           hwloc_obj_type_snprintf(type, sizeof(type), nodes[i], 1);
           idx = logicalo ? nodes[i]->logical_index : nodes[i]->os_index;
-          printf("%s%u", i==0 ? (const char *) "" : sep, idx);
+          if (first)
+            first = 0;
+          else
+            printf("%s", sep);
+          if (objecto) {
+            char types[64];
+            hwloc_obj_type_snprintf(types, sizeof(types), nodes[i], 0);
+            printf("%s:", types);
+          }
+          printf("%u", idx);
         }
       }
-      free(nodes);
     }
+    free(nodes);
+    hwloc_bitmap_free(nodeset);
     printf("\n");
 
   } else {
     char *string = NULL;
-    if (taskset)
-      hwloc_bitmap_taskset_asprintf(&string, set);
-    else
-      hwloc_bitmap_asprintf(&string, set);
+    hwloc_utils_cpuset_format_asprintf(&string, set, cpuset_output_format);
     printf("%s\n", string);
     free(string);
   }
   return EXIT_SUCCESS;
 }
 
-static int hwloc_calc_type_depth(hwloc_topology_t topology, const char *string, int *depthp, union hwloc_obj_attr_u *attrp, const char *caller)
-{
-  union hwloc_obj_attr_u attr;
-  hwloc_obj_type_t type;
-  int depth;
-  int err;
-
-  /* similar to hwloc_type_sscanf_as_depth() but we want to get attr as well */
-
-  err = hwloc_type_sscanf(string, &type, &attr, sizeof(attr));
-  if (err < 0) {
-    char *endptr;
-    depth = strtoul(string, &endptr, 0);
-    if (*endptr) {
-      fprintf(stderr, "unrecognized %s type or depth %s\n", caller, string);
-      return -1;
-    }
-
-    *depthp = depth;
-    return 0;
-  }
-
-  depth = hwloc_get_type_depth(topology, type);
-  if (type == HWLOC_OBJ_GROUP
-      && depth == HWLOC_TYPE_DEPTH_MULTIPLE
-      && attr.group.depth != (unsigned)-1) {
-    unsigned l;
-    depth = HWLOC_TYPE_DEPTH_UNKNOWN;
-    for(l=0; l<(unsigned) hwloc_topology_get_depth(topology); l++) {
-      hwloc_obj_t tmp = hwloc_get_obj_by_depth(topology, l, 0);
-      if (tmp->type == HWLOC_OBJ_GROUP && tmp->attr->group.depth == attr.group.depth) {
-	depth = (int)l;
-	break;
-      }
-    }
-  }
-
-  if (depth == HWLOC_TYPE_DEPTH_UNKNOWN) {
-    fprintf(stderr, "unavailable %s type %s\n", caller, hwloc_obj_type_string(type));
-    return -1;
-  } else  if (depth == HWLOC_TYPE_DEPTH_MULTIPLE) {
-    fprintf(stderr, "cannot use %s type %s with multiple depth, please use the relevant depth\n", caller, hwloc_obj_type_string(type));
-    return -1;
-  }
-
-  if (attrp)
-    memcpy(attrp, &attr, sizeof(attr));
-  *depthp = depth;
-  return 0;
-}
-
 int main(int argc, char *argv[])
 {
   hwloc_topology_t topology;
   unsigned long flags = HWLOC_TOPOLOGY_FLAG_IMPORT_SUPPORT;
   unsigned long restrict_flags = 0;
   char *input = NULL;
-  enum hwloc_utils_input_format input_format = HWLOC_UTILS_INPUT_DEFAULT;
+  struct hwloc_utils_input_format_s input_format = HWLOC_UTILS_INPUT_FORMAT_DEFAULT;
   int depth = 0;
   hwloc_bitmap_t set;
   int cmdline_args = 0;
-  const char * numberoftype = NULL;
-  const char * intersecttype = NULL;
+  const char * numberof_string = NULL;
+  const char * intersect_string = NULL;
   char *restrictstring = NULL;
-  char * hiertype = NULL;
+  char * hier_string = NULL;
   char * best_memattr_str = NULL;
   char *callname;
   char *outsep = NULL;
   int cpukind_index = -1;
   char *cpukind_infoname = NULL;
   char *cpukind_infovalue = NULL;
+  enum hwloc_utils_cpuset_format_e cpuset_input_format = HWLOC_UTILS_CPUSET_FORMAT_UNKNOWN;
   int opt;
   int i;
   int err;
@@ -413,7 +376,7 @@ int main(int argc, char *argv[])
       if (equal) {
         cpukind_infoname = argv[1];
         cpukind_infovalue = equal+1;
-        *equal = 0;
+        *equal = '\0';
       } else if (argv[1][0] >= '0' && argv[1][0] <= '9') {
         cpukind_index = atoi(argv[1]);
       } else {
@@ -486,6 +449,9 @@ int main(int argc, char *argv[])
       /* FALLTHRU */
     }
   }
+  if (input) {
+    hwloc_utils_disable_input_format(&input_format);
+  }
 
   while (argc >= 1) {
     opt = 0;
@@ -520,7 +486,7 @@ int main(int argc, char *argv[])
 	  usage(callname, stderr);
 	  return EXIT_FAILURE;
 	}
-	numberoftype = argv[1];
+	numberof_string = argv[1];
 	opt = 1;
 	goto next;
       }
@@ -529,7 +495,7 @@ int main(int argc, char *argv[])
 	  usage(callname, stderr);
 	  return EXIT_FAILURE;
 	}
-	intersecttype = argv[1];
+	intersect_string = argv[1];
 	opt = 1;
 	goto next;
       }
@@ -538,7 +504,7 @@ int main(int argc, char *argv[])
 	  usage(callname, stderr);
 	  return EXIT_FAILURE;
 	}
-	hiertype = argv[1];
+	hier_string = argv[1];
 	opt = 1;
 	goto next;
       }
@@ -568,12 +534,12 @@ int main(int argc, char *argv[])
       }
       if (!strcasecmp(argv[0], "--pulist") || !strcmp(argv[0], "--proclist")) {
 	/* backward compat with 1.0 */
-	intersecttype = "pu";
+	intersect_string = "pu";
         goto next;
       }
       if (!strcmp(argv[0], "--nodelist")) {
 	/* backward compat with 1.0 */
-	intersecttype = "numa";
+	intersect_string = "numa";
         goto next;
       }
       if (!strcmp(argv[0], "--largest")  || !strcmp(argv[0], "--objects") /* backward compat with 1.0 */) {
@@ -623,6 +589,10 @@ int main(int argc, char *argv[])
 	nodeseto = 1;
 	goto next;
       }
+      if (!strcmp(argv[0], "--oo") || !strcmp(argv[0], "--object-output")) {
+	objecto = 1;
+	goto next;
+      }
       if (!strcmp(argv[0], "--sep")) {
 	if (argc < 2) {
 	  usage (callname, stderr);
@@ -636,10 +606,40 @@ int main(int argc, char *argv[])
 	singlify = 1;
 	goto next;
       }
-      if (!strcmp(argv[0], "--taskset")) {
-	taskset = 1;
+      if (!strcmp(argv[0], "--cpuset-output-format") || !strcmp(argv[0], "--cof")) {
+	if (argc < 2) {
+	  usage (callname, stderr);
+	  exit(EXIT_FAILURE);
+	}
+        cpuset_output_format = hwloc_utils_parse_cpuset_format(argv[1]);
+        if (HWLOC_UTILS_CPUSET_FORMAT_UNKNOWN == cpuset_output_format) {
+          fprintf(stderr, "Unrecognized %s argument %s\n", argv[0], argv[1]);
+          exit(EXIT_FAILURE);
+        }
+	opt = 1;
 	goto next;
       }
+      if (!strcmp(argv[0], "--cpuset-input-format") || !strcmp(argv[0], "--cif")) {
+	if (argc < 2) {
+	  usage (callname, stderr);
+	  exit(EXIT_FAILURE);
+	}
+        cpuset_input_format = hwloc_utils_parse_cpuset_format(argv[1]);
+        if (HWLOC_UTILS_CPUSET_FORMAT_UNKNOWN == cpuset_input_format) {
+          fprintf(stderr, "Unrecognized %s argument %s\n", argv[0], argv[1]);
+          exit(EXIT_FAILURE);
+        }
+        if (HWLOC_UTILS_CPUSET_FORMAT_SYSTEMD == cpuset_input_format) {
+          fprintf(stderr, "The systemd-dbus-api format is not supported on input\n");
+          exit(EXIT_FAILURE);
+        }
+	opt = 1;
+	goto next;
+      }
+      if (!strcmp(argv[0], "--taskset")) {
+        cpuset_output_format = HWLOC_UTILS_CPUSET_FORMAT_TASKSET;
+        goto next;
+      }
 
       fprintf (stderr, "Unrecognized option: %s\n", argv[0]);
       usage(callname, stderr);
@@ -655,6 +655,7 @@ int main(int argc, char *argv[])
     scontext.output_set = set;
     scontext.nodeset_input = nodeseti;
     scontext.nodeset_output = nodeseto;
+    scontext.cpuset_input_format = cpuset_input_format;
     if (hwloc_calc_process_location_as_set(&lcontext, &scontext, argv[0]) < 0)
       fprintf(stderr, "ignored unrecognized argument %s\n", argv[0]);
 
@@ -669,16 +670,34 @@ int main(int argc, char *argv[])
     argv += opt+1;
   }
 
-  if (numberoftype && hwloc_calc_type_depth(topology, numberoftype, &numberofdepth, &numberofattr, "--number-of") < 0)
+  numberof.depth = HWLOC_TYPE_DEPTH_UNKNOWN; /* disable this feature by default */
+  if (numberof_string && hwloc_calc_parse_level(NULL, topology, numberof_string, strlen(numberof_string), &numberof) < 0) {
+    if (numberof.depth == HWLOC_TYPE_DEPTH_MULTIPLE)
+      fprintf(stderr, "cannot use --number-of type %s with multiple depth, please use the relevant depth\n",
+              numberof_string);
+    else if (numberof.depth == HWLOC_TYPE_DEPTH_UNKNOWN)
+      fprintf(stderr, "cannot use --number-of type %s, unavailable\n",
+              numberof_string);
     goto out;
+  }
 
-  if (intersecttype && hwloc_calc_type_depth(topology, intersecttype, &intersectdepth, &intersectattr, "--intersect") < 0)
+  intersect.depth = HWLOC_TYPE_DEPTH_UNKNOWN; /* disable this feature by default */
+  if (intersect_string && hwloc_calc_parse_level(NULL, topology, intersect_string, strlen(intersect_string), &intersect) < 0) {
+    if (intersect.depth == HWLOC_TYPE_DEPTH_MULTIPLE)
+      fprintf(stderr, "cannot use --intersect type %s with multiple depth, please use the relevant depth\n",
+              intersect_string);
+    else if (intersect.depth == HWLOC_TYPE_DEPTH_UNKNOWN)
+      fprintf(stderr, "cannot use --intersect type %s, unavailable\n",
+              intersect_string);
     goto out;
+  }
 
-  if (hiertype) {
+  hiernblevels = 0; /* disable this feature by default */
+  hierlevels = NULL;
+  if (hier_string) {
     char *tmp, *next;
     hiernblevels = 1;
-    tmp = hiertype;
+    tmp = hier_string;
     while (1) {
       tmp = strchr(tmp, '.');
       if (!tmp)
@@ -686,15 +705,22 @@ int main(int argc, char *argv[])
       tmp++;
       hiernblevels++;
     }
-    hierdepth = malloc(hiernblevels * sizeof(int));
-    tmp = hiertype;
+    hierlevels = malloc(hiernblevels * sizeof(struct hwloc_calc_level));
+    tmp = hier_string;
     for(i=0; i<hiernblevels; i++) {
       next = strchr(tmp, '.');
       if (next)
 	*next = '\0';
-      if (hwloc_calc_type_depth(topology, tmp, &hierdepth[i], NULL, "--hierarchical") < 0)
+      if (hwloc_calc_parse_level(NULL, topology, tmp, strlen(tmp), &hierlevels[i]) < 0) {
+        if (hierlevels[i].depth == HWLOC_TYPE_DEPTH_MULTIPLE)
+          fprintf(stderr, "cannot use --hierarchical %s with multiple depth, please use the relevant depth\n",
+                  tmp);
+        else if (hierlevels[i].depth == HWLOC_TYPE_DEPTH_UNKNOWN)
+          fprintf(stderr, "cannot use --hierarchical type %s, unavailable\n",
+                  tmp);
 	goto out;
-      if (hierdepth[i] < 0 && hierdepth[i] != HWLOC_TYPE_DEPTH_NUMANODE) {
+      }
+      if (hierlevels[i].depth < 0 && hierlevels[i].depth != HWLOC_TYPE_DEPTH_NUMANODE) {
 	fprintf(stderr, "unsupported (non-normal) --hierarchical type %s\n", tmp);
 	goto out;
       }
@@ -703,6 +729,8 @@ int main(int argc, char *argv[])
   }
 
   if (best_memattr_str) {
+    best_node_flags = hwloc_utils_parse_best_node_flags(best_memattr_str);
+
     best_memattr_id = hwloc_utils_parse_memattr_name(topology, best_memattr_str);
     if (best_memattr_id == (hwloc_memattr_id_t) -1) {
       fprintf(stderr, "unrecognized memattr %s\n", best_memattr_str);
@@ -719,7 +747,8 @@ int main(int argc, char *argv[])
 #define HWLOC_CALC_LINE_LEN 64
     size_t len = HWLOC_CALC_LINE_LEN;
     char * line = malloc(len);
-    printf("Waiting for locations to process on stdin...\n");
+    if (verbose >= 0)
+      printf("Waiting for locations to process on stdin...\n");
     while (1) {
       char *current, *tmpline;
 
@@ -760,6 +789,7 @@ int main(int argc, char *argv[])
 	scontext.output_set = set;
 	scontext.nodeset_input = nodeseti;
 	scontext.nodeset_output = nodeseto;
+        scontext.cpuset_input_format = cpuset_input_format;
 	if (hwloc_calc_process_location_as_set(&lcontext, &scontext, token) < 0)
 	  fprintf(stderr, "ignored unrecognized argument %s\n", token);
       }
@@ -774,7 +804,7 @@ int main(int argc, char *argv[])
   hwloc_bitmap_free(set);
   hwloc_bitmap_free(cpukind_cpuset);
 
-  free(hierdepth);
+  free(hierlevels);
 
   return ret;
 }
